//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://emule.sf.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "downloadqueue.h"
#include "updownclient.h"
#include "partfile.h"

CDownloadQueue::CDownloadQueue(CPreferences* in_prefs,CSharedFileList* in_sharedfilelist){
	app_prefs = in_prefs;
	sharedfilelist = in_sharedfilelist;
	filesrdy = 0;
	datarate = 0;
}

void CDownloadQueue::Init(){
	// find all part files, read & hash them if needed and store into a list
	CFileFind ff;
	char* searchpath = new char[strlen(app_prefs->GetTempDir())+15];
	sprintf(searchpath,"%s\\*.part.met",app_prefs->GetTempDir());
	bool end = !ff.FindFile(searchpath,0);
	delete[] searchpath;
	if (end){
		theApp.emuledlg->AddLogLine(false,"No part files found");
		return;
	}
	
	uint16 count = 0;
	while (!end){
		end = !ff.FindNextFile();
		if (ff.IsDirectory())
			continue;
		CPartFile* toadd = new CPartFile();
		if (toadd->LoadPartFile(app_prefs->GetTempDir(),ff.GetFileName().GetBuffer())){
			count++;
			filelist.AddTail(toadd);			// to downloadqueue
			if (toadd->GetStatus() == PS_READY)
				sharedfilelist->SafeAddKFile(toadd); // part files are always shared files
			theApp.emuledlg->transferwnd.downloadlistctrl.AddFile(toadd);// show in downloadwindow
		}
		else
			delete toadd;
	}
	ff.Close();
	theApp.emuledlg->AddLogLine(false,"Found %i part files",count);
}

CDownloadQueue::~CDownloadQueue(){
	for (POSITION pos =filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		filelist.GetAt(pos)->SavePartFile();
		delete filelist.GetAt(pos);
	}
	
}

void CDownloadQueue::AddSearchToDownload(CSearchFile* toadd){
	if (CKnownFile* file = sharedfilelist->GetFileByID(toadd->GetFileHash())){
		if (file->IsPartFile())
			MessageBox(0,"You are already downloading this file",file->GetFileName(),64);
		else
			MessageBox(0,"You already downloaded this file",file->GetFileName(),64);
		return;
	}
	else if ( file = this->GetFileByID(toadd->GetFileHash())){
		MessageBox(0,"You are already downloading this file",file->GetFileName(),64);
		return;
	}
	CPartFile* newfile = new CPartFile(toadd);
	filelist.AddTail(newfile);
	theApp.emuledlg->transferwnd.downloadlistctrl.AddFile(newfile);
	theApp.emuledlg->AddLogLine(true,"Downloading %s",newfile->GetFileName());
}

void CDownloadQueue::Process(){
	sint16 downspeed = 0;
	if (app_prefs->GetMaxDownload() != UNLIMITED){
		downspeed =  (uint16)((float)((float)(app_prefs->GetMaxDownload()*1024)/(datarate+1)) * 100);
		if (downspeed < 70)
			downspeed = 70;
		else if (downspeed > 200)
			downspeed = 200;
	}
	datarate = 0;
	for (POSITION pos =filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		CPartFile* cur_file =  filelist.GetAt(pos);
		if (cur_file->GetStatus() == PS_READY || cur_file->GetStatus() == PS_EMPTY){
			datarate += cur_file->Process(downspeed);
		}
	}	
}

CPartFile*	CDownloadQueue::GetFileByID(uchar* filehash){
	for (POSITION pos = filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		if (!memcmp(filehash,filelist.GetAt(pos)->GetFileHash(),16))
			return filelist.GetAt(pos);
	}
	return 0;
}

bool CDownloadQueue::IsPartFile(void* totest){
	for (POSITION pos = filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos))
		if (totest == filelist.GetAt(pos))
			return true;
	return false;
}

void CDownloadQueue::CheckAndAddSource(CPartFile* sender,CUpDownClient* source){
	for (POSITION pos = filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		CPartFile* cur_file = filelist.GetAt(pos);
		for (POSITION pos2 = cur_file->srclist.GetHeadPosition();pos2 != 0; cur_file->srclist.GetNext(pos2))
			if (cur_file->srclist.GetAt(pos2)->Compare(source)){
				if (cur_file == sender){ // this file has already this source
					delete source;
					return;
				}
				// set request for this source
				if (cur_file->srclist.GetAt(pos2)->AddRequestForAnotherFile(sender)){
					// add it to uploadlistctrl
					theApp.emuledlg->transferwnd.downloadlistctrl.AddSource(sender,cur_file->srclist.GetAt(pos2),true);
					delete source;
					return;
				}
				else{
					delete source;
					return;
				}
			}
	}
	//our new source is real new but maybe it is already uploading to us?
	//if yes the known client will be attached to the var "source"
	//and the old sourceclient will be deleted
	theApp.uploadqueue->AttachToAlreadyKnown(&source,0);
	sender->srclist.AddTail(source);
	theApp.emuledlg->transferwnd.downloadlistctrl.AddSource(sender,source,false);
	theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this);
}

bool CDownloadQueue::AttachToAlreadyKnown(CUpDownClient** client, CClientReqSocket* sender){
	for (POSITION pos = filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		CPartFile* cur_file = filelist.GetAt(pos);
		for (POSITION pos2 = cur_file->srclist.GetHeadPosition();pos2 != 0; cur_file->srclist.GetNext(pos2)){
			if ((*client)->Compare(cur_file->srclist.GetAt(pos2))){
				if (sender){
					safe_delete(cur_file->srclist.GetAt(pos2)->socket);
					cur_file->srclist.GetAt(pos2)->socket = sender;
					(*client)->socket = 0;
				}
				delete *client;
				*client = cur_file->srclist.GetAt(pos2);
				return true;
			}
		}
	}
	return false;
}

bool CDownloadQueue::RemoveSource(CUpDownClient* toremove, bool	updatewindow){
	bool removed = false;
	for (POSITION pos = filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		CPartFile* cur_file = filelist.GetAt(pos);
		for (POSITION pos2 = cur_file->srclist.GetHeadPosition();pos2 != 0; cur_file->srclist.GetNext(pos2)){
			if (toremove == cur_file->srclist.GetAt(pos2)){
				cur_file->srclist.RemoveAt(pos2);
				removed = true;
				break;
			}
		}
	}
	if (updatewindow){
		theApp.emuledlg->transferwnd.downloadlistctrl.RemoveSource(toremove,0);
		toremove->SetDownloadState(DS_NONE);
	}
	return removed;
}

void CDownloadQueue::RemoveFile(CPartFile* toremove){
	for (POSITION pos = filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		if (toremove == filelist.GetAt(pos)){
			filelist.RemoveAt(pos);
			return;
		}
	}
}

void CDownloadQueue::DeleteAll(){
	for (POSITION pos = filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		CPartFile* cur_file = filelist.GetAt(pos);
		for (POSITION pos2 = cur_file->srclist.GetHeadPosition();pos2 != 0; cur_file->srclist.GetNext(pos2)){	
			// to make sure there is no invalid client in there
			theApp.uploadqueue->RemoveFromUploadQueue(cur_file->srclist.GetAt(pos2),false);
			theApp.uploadqueue->RemoveFromWaitingQueue(cur_file->srclist.GetAt(pos2),false);
			// now delete
			delete cur_file->srclist.GetAt(pos2);
		}
		cur_file->srclist.RemoveAll();
	}
}